from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import (
    icon_acl_delete, icon_acl_list, icon_acl_create, icon_acl_permissions,
    icon_global_acl_list
)
from .permissions import permission_acl_view, permission_acl_edit


link_acl_create = Link(
    icon=icon_acl_create, kwargs=get_content_type_kwargs_factory(
        variable_name='resolved_object'
    ), permissions=(permission_acl_edit,), text=_('New ACL'),
    view='acls:acl_create'
)
link_acl_delete = Link(
    args='resolved_object.pk', icon=icon_acl_delete,
    permissions=(permission_acl_edit,), tags='dangerous', text=_('Delete'),
    view='acls:acl_delete'
)
link_acl_list = Link(
    icon=icon_acl_list, kwargs=get_content_type_kwargs_factory(
        variable_name='resolved_object'
    ), permissions=(permission_acl_view,), text=_('ACLs'), view='acls:acl_list'
)
link_acl_permissions = Link(
    args='resolved_object.pk', icon=icon_acl_permissions,
    permissions=(permission_acl_edit,),
    text=_('Permissions'), view='acls:acl_permissions'
)
link_global_acl_list = Link(
    icon=icon_global_acl_list, text=_('Global ACLs'),
    view='acls:global_acl_list'
)



from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.permissions.serializers import PermissionSerializer
from mayan.apps.rest_api import serializers

from .models import Source


class SourceBackendActionInterfaceArgumentSerializer(serializers.Serializer):
    default = serializers.SerializerMethodField(
        label=_('Default')
    )
    has_default = serializers.BooleanField(
        label=_('Has default?')
    )
    help_text = serializers.CharField(
        label=_('Help text')
    )
    name = serializers.CharField(
        label=_('Name')
    )
    required = serializers.BooleanField(
        label=_('Required')
    )

    def get_default(self, instance):
        if instance.has_default:
            return instance.default
        else:
            return


class SourceBackendActionInterfaceSerializer(serializers.Serializer):
    arguments = SourceBackendActionInterfaceArgumentSerializer(
        label=_('Arguments'), many=True, read_only=True,
        source='arguments_visible'
    )
    name = serializers.CharField(
        label=_('Name')
    )


class SourceBackendActionSerializer(serializers.Serializer):
    accept_files = serializers.BooleanField(
        label=_('Accept files'), read_only=True
    )
    arguments = serializers.JSONField(
        help_text=_(
            'Optional arguments for the action. Must be JSON formatted.'
        ), initial={}, label=_('Arguments'), required=False,
        write_only=True
    )
    confirmation = serializers.BooleanField(
        label=_('Confirmation'), read_only=True
    )
    execute_url = serializers.SerializerMethodField(
        label=_('Execute URL'), read_only=True
    )
    interfaces = SourceBackendActionInterfaceSerializer(
        label=_('Interfaces'), many=True, read_only=True,
        source='interface_visible_list'
    )
    name = serializers.CharField(
        label=_('name'), read_only=True
    )
    permission = PermissionSerializer(
        label=_('Permission'), read_only=True
    )
    source_id = serializers.SerializerMethodField()
    source_url = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        action = self.context.get('action')

        if action:
            if action.accept_files:
                self.fields['file'] = serializers.FileField(
                    help_text=_('Binary content for the new file.'),
                    use_url=False, write_only=True
                )

    def get_execute_url(self, instance):
        return reverse(
            viewname='rest_api:source_action-execute', kwargs={
                'source_id': instance.source.pk,
                'action_name': instance.name
            }, request=self.context['request'],
            format=self.context['format']
        )

    def get_source_id(self, instance):
        return instance.source.pk

    def get_source_url(self, instance):
        return reverse(
            viewname='rest_api:source-detail', kwargs={
                'source_id': instance.source.pk,
            }, request=self.context['request'],
            format=self.context['format']
        )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:source_action-detail', kwargs={
                'source_id': instance.source.pk,
                'action_name': instance.name
            }, request=self.context['request'],
            format=self.context['format']
        )


class SourceSerializer(serializers.HyperlinkedModelSerializer):
    actions_url = serializers.SerializerMethodField(
        label=_('Actions URL')
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_field': 'pk', 'lookup_url_kwarg': 'source_id',
                'view_name': 'rest_api:source-detail'
            },
        }
        fields = (
            'actions_url', 'backend_data', 'backend_path', 'enabled', 'id',
            'label', 'url'
        )
        model = Source
        read_only_fields = ('actions', 'id', 'url')

    def get_actions_url(self, instance):
        return reverse(
            viewname='rest_api:source_action-list', kwargs={
                'source_id': instance.pk,
            }, request=self.context['request'],
            format=self.context['format']
        )



from unittest import mock

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_version_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins import EmailSourceTestMixin


class SourceBackendDocumentUploadCallbackTestCase(
    EmailSourceTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_create'
    )
    def test_callback_post_document_create_on_document_upload(
        self, mocked_source_method
    ):
        self._silence_logger(name='mayan.apps.converter')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_file_create'
    )
    def test_callback_post_document_file_create_on_document_create(
        self, mocked_source_method
    ):
        self._silence_logger(name='mayan.apps.converter')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_file_upload'
    )
    def test_callback_post_document_file_upload_on_document_upload(
        self, mocked_source_method
    ):
        self._silence_logger(name='mayan.apps.converter')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from mayan.apps.converter.layers import layer_saved_transformations

from ...document_version_modifications import (
    DocumentVersionModificationPagesAppend,
    DocumentVersionModificationPagesReset
)
from ...literals import PAGE_RANGE_ALL
from ...models.document_version_models import DocumentVersion
from ...models.document_version_page_models import DocumentVersionPage

from ..literals import (
    TEST_DOCUMENT_VERSION_COMMENT_EDITED, TEST_TRANSFORMATION_ARGUMENT,
    TEST_TRANSFORMATION_CLASS
)

from .document_mixins import DocumentTestMixin


class DocumentVersionAPIViewTestMixin:
    def _request_test_document_version_create_api_view(self):
        pk_list = list(
            DocumentVersion.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:documentversion-list', kwargs={
                'document_id': self._test_document.pk
            }
        )

        try:
            self._test_document_version = DocumentVersion.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentVersion.DoesNotExist:
            self._test_document_version = None

        return response

    def _request_test_document_version_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documentversion-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document.version_active.pk
            }
        )

    def _request_test_document_version_detail_api_view(self):
        return self.get(
            viewname='rest_api:documentversion-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document.version_active.pk
            }
        )

    def _request_test_document_version_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documentversion-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document.version_active.pk
            }, data={'comment': TEST_DOCUMENT_VERSION_COMMENT_EDITED}
        )

    def _request_test_document_version_edit_via_put_api_view(
        self, extra_view_kwargs=None
    ):
        view_kwargs = {
            'document_id': self._test_document.pk,
            'document_version_id': self._test_document.version_active.pk
        }

        if extra_view_kwargs:
            view_kwargs.update(**extra_view_kwargs)

        return self.put(
            viewname='rest_api:documentversion-detail', kwargs=view_kwargs,
            data={
                'active': True,
                'comment': TEST_DOCUMENT_VERSION_COMMENT_EDITED
            }
        )

    def _request_test_document_version_list_api_view(self):
        return self.get(
            viewname='rest_api:documentversion-list', kwargs={
                'document_id': self._test_document.pk
            }
        )


class DocumentVersionLinkTestMixin:
    def _resolve_test_document_version_link(self, test_link):
        self.add_test_view(test_object=self._test_document_version)
        context = self.get_test_view()
        return test_link.resolve(context=context)


class DocumentVersionModificationAPIViewTestMixin:
    def _request_test_document_version_action_page_append_api_view(self):
        return self.post(
            viewname='rest_api:documentversion-modify', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk
            }, data={
                'backend_id': DocumentVersionModificationPagesAppend.backend_id
            }
        )

    def _request_test_document_version_action_page_reset_api_view(self):
        return self.post(
            viewname='rest_api:documentversion-modify', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk
            }, data={
                'backend_id': DocumentVersionModificationPagesReset.backend_id
            }
        )


class DocumentVersionModificationViewTestMixin:
    def _request_test_document_version_action_page_append_view(self):
        return self.post(
            viewname='documents:document_version_modify', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'backend': DocumentVersionModificationPagesAppend.backend_id
            }
        )

    def _request_test_document_version_action_page_reset_view(self):
        return self.post(
            viewname='documents:document_version_modify', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'backend': DocumentVersionModificationPagesReset.backend_id
            }
        )


class DocumentVersionPageAPIViewTestMixin:
    def _request_test_document_version_page_create_api_view(self):
        pk_list = list(
            DocumentVersionPage.objects.values_list('pk', flat=True)
        )

        content_type = ContentType.objects.get_for_model(
            model=self._test_document_file_page
        )

        response = self.post(
            viewname='rest_api:documentversionpage-list', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk
            }, data={
                'content_type_id': content_type.pk,
                'object_id': self._test_document_file_page.pk,
                'page_number': self._test_document_file_page.page_number + 2
            }
        )

        try:
            self._test_document_version_page = DocumentVersionPage.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentVersionPage.DoesNotExist:
            self._test_document_version_page = None

        return response

    def _request_test_document_version_page_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documentversionpage-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_detail_api_view(self):
        return self.get(
            viewname='rest_api:documentversionpage-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documentversionpage-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }, data={
                'content_type_id': self._test_document_version_page.content_type.pk,
                'page_number': self._test_document_version_page.page_number + 1
            }
        )

    def _request_test_document_version_page_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:documentversionpage-detail', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }, data={
                'content_type_id': self._test_document_version_page.content_type.pk,
                'object_id': self._test_document_version_page.object_id,
                'page_number': self._test_document_version_page.page_number + 1
            }
        )

    def _request_test_document_version_page_image_api_view(self):
        return self.get(
            viewname='rest_api:documentversionpage-image', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_list_api_view(self):
        return self.get(
            viewname='rest_api:documentversionpage-list', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
            }
        )


class DocumentVersionTestMixin(DocumentTestMixin):
    auto_create_test_document_version = False
    auto_create_test_document_version_page = False

    def setUp(self):
        super().setUp()

        if self.auto_create_test_document_version:
            self._create_test_document_version()

        if self.auto_create_test_document_version_page:
            self._create_test_document_version_page()

    def _create_test_document_version(self, user=None):
        if self._test_document.versions.count() == 0:
            active = True
        else:
            active = False

        self._test_document_version = DocumentVersion(
            active=active, document=self._test_document
        )
        self._test_document_version._event_actor = user
        self._test_document_version.save()

        self._test_document_version_list.append(self._test_document_version)

    def _create_test_document_version_page(self):
        self._document_version_page_source_object = self._test_document

        content_type = ContentType.objects.get_for_model(
            model=self._document_version_page_source_object
        )

        self._test_document_version_page = DocumentVersionPage.objects.create(
            content_type=content_type,
            document_version=self._test_document_version,
            object_id=self._document_version_page_source_object.pk
        )


class DocumentVersionViewTestMixin:
    def _request_test_document_version_active_view(self):
        return self.post(
            viewname='documents:document_version_active', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_single_delete_view(self):
        return self.post(
            viewname='documents:document_version_single_delete', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_multiple_delete_view(self):
        return self.post(
            viewname='documents:document_version_multiple_delete', data={
                'id_list': self._test_document_version.pk
            }
        )

    def _request_test_document_version_edit_view(self):
        return self.post(
            viewname='documents:document_version_edit', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'comment': TEST_DOCUMENT_VERSION_COMMENT_EDITED
            }
        )

    def _request_test_document_version_list_view(self):
        return self.get(
            viewname='documents:document_version_list', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_version_preview_view(self):
        return self.get(
            viewname='documents:document_version_preview', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_print_form_view(self):
        return self.get(
            viewname='documents:document_version_print_form', kwargs={
                'document_version_id': self._test_document_version.pk,
            }, data={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_version_print_view(self):
        return self.get(
            viewname='documents:document_version_print_view', kwargs={
                'document_version_id': self._test_document_version.pk,
            }, query={
                'page_group': PAGE_RANGE_ALL
            }
        )


class DocumentVersionPageViewTestMixin:
    def _request_test_document_version_page_delete_view(self):
        return self.post(
            viewname='documents:document_version_page_delete', kwargs={
                'document_version_page_id': self._test_document_version_page.pk,
            }
        )

    def _request_test_document_version_page_list_view(self):
        return self.get(
            viewname='documents:document_version_page_list', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_page_rotate_left_view(self):
        return self.post(
            viewname='documents:document_version_page_rotate_left', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_rotate_right_view(self):
        return self.post(
            viewname='documents:document_version_page_rotate_right', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_view(self, document_version_page):
        return self.get(
            viewname='documents:document_version_page_view', kwargs={
                'document_version_page_id': document_version_page.pk,
            }
        )

    def _request_test_document_version_page_zoom_in_view(self):
        return self.post(
            viewname='documents:document_version_page_zoom_in', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_zoom_out_view(self):
        return self.post(
            viewname='documents:document_version_page_zoom_out', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }
        )


class DocumentVersionPageRemapViewTestMixin:
    def _request_test_document_version_page_list_remap_view(self, data):
        return self.post(
            viewname='documents:document_version_page_list_remap', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data=data
        )


class DocumentVersionTransformationTestMixin:
    def _create_document_version_transformation(self):
        layer_saved_transformations.add_transformation_to(
            obj=self._test_document_version.pages.first(),
            transformation_class=TEST_TRANSFORMATION_CLASS,
            arguments=TEST_TRANSFORMATION_ARGUMENT
        )


class DocumentVersionTransformationViewTestMixin:
    def _request_test_document_version_transformations_clear_view(self):
        return self.post(
            viewname='documents:document_version_transformations_clear',
            kwargs={'document_version_id': self._test_document_version.pk}
        )

    def _request_test_document_version_multiple_transformations_clear_view(self):
        return self.post(
            viewname='documents:document_version_multiple_transformations_clear',
            data={'id_list': self._test_document_version.pk}
        )

    def _request_test_document_version_transformations_clone_view(self):
        return self.post(
            viewname='documents:document_version_transformations_clone',
            kwargs={'document_version_id': self._test_document_version.pk}, data={
                'page': self._test_document_version.pages.first().pk
            }
        )



from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.validators import validate_internal_name
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from ..events import (
    event_workflow_template_created, event_workflow_template_edited
)
from ..managers import WorkflowManager

from .workflow_model_mixins import WorkflowBusinessLogicMixin

__all__ = ('Workflow', 'WorkflowRuntimeProxy')


class Workflow(
    ExtraDataModelMixin, WorkflowBusinessLogicMixin, models.Model
):
    """
    Fields:
    * label - Identifier. A name/label to call the workflow
    """
    auto_launch = models.BooleanField(
        default=True, help_text=_(
            'Launch workflow when document is created.'
        ), verbose_name=_('Auto launch')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used by other apps to reference this '
            'workflow. Can only contain letters, numbers, and underscores.'
        ), max_length=255, unique=True, validators=[validate_internal_name],
        verbose_name=_('Internal name')
    )
    label = models.CharField(
        help_text=_('A short text to describe the workflow.'),
        max_length=255, unique=True, verbose_name=_('Label')
    )
    document_types = models.ManyToManyField(
        related_name='workflows', to=DocumentType, verbose_name=_(
            'Document types'
        )
    )

    objects = WorkflowManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Workflow')
        verbose_name_plural = _('Workflows')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='document_states:workflow_template_list')

    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_workflow_template_created,
            'target': 'self'
        },
        edited={
            'event': event_workflow_template_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class WorkflowRuntimeProxy(Workflow):
    class Meta:
        proxy = True
        verbose_name = _('Workflow runtime proxy')
        verbose_name_plural = _('Workflow runtime proxies')

    def get_documents(self, permission=None, user=None):
        """
        Provide a queryset of the documents. The queryset is optionally
        filtered by access.
        """
        queryset = Document.valid.filter(workflows__workflow=self)

        if permission and user:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=user
            )

        return queryset

    def get_document_count(self, user):
        """
        Return the numeric count of documents executing this workflow.
        The count is filtered by access.
        """
        return self.get_documents(
            permission=permission_document_view, user=user
        ).count()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.credentials.models import StoredCredential
from mayan.apps.credentials.permissions import permission_credential_use

from .classes import MailerBackendBaseEmail


class DjangoSMTP(MailerBackendBaseEmail):
    """
    Backend that wraps Django's SMTP backend
    """
    class_path = 'django.core.mail.backends.smtp.EmailBackend'
    label = _('Django SMTP backend')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'host': {
                    'label': _('Host'),
                    'class': 'django.forms.CharField',
                    'default': 'localhost',
                    'help_text': _('The host to use for sending email.'),
                    'kwargs': {
                        'max_length': 48
                    }, 'required': True
                }, 'port': {
                    'label': _('Port'),
                    'class': 'django.forms.IntegerField', 'default': 25,
                    'help_text': _('Port to use for the SMTP server.'),
                    'required': True
                }, 'use_tls': {
                    'label': _('Use TLS'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        'Whether to use a TLS (secure) connection when '
                        'talking to the SMTP server. This is used for '
                        'explicit TLS connections, generally on port 587.'
                    ), 'required': False
                }, 'use_ssl': {
                    'label': _('Use SSL'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        'Whether to use an implicit TLS (secure) connection '
                        'when talking to the SMTP server. In most email '
                        'documentation this type of TLS connection is '
                        'referred to as SSL. It is generally used on port '
                        '465. If you are experiencing problems, see the '
                        'explicit TLS setting "Use TLS". Note that '
                        '"Use TLS" and "Use SSL" are mutually exclusive, '
                        'so only set one of those settings to True.'
                    ), 'required': False
                }, 'stored_credential_id': {
                    'class': 'mayan.apps.views.fields.FormFieldFilteredModelChoice',
                    'help_text': _(
                        'The credential entry to use to authenticate '
                        'against the email server.'
                    ),
                    'kwargs': {
                        'source_model': StoredCredential,
                        'permission': permission_credential_use
                    },
                    'label': _('Credential'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Server'), {
                    'fields': ('host', 'port', 'use_tls', 'use_ssl')
                }
            ), (
                _('Authentication'), {
                    'fields': ('stored_credential_id',)
                }
            )
        )

        return fieldsets


class DjangoFileBased(MailerBackendBaseEmail):
    """
    Mailing backend that wraps Django's file based email backend
    """
    class_path = 'django.core.mail.backends.filebased.EmailBackend'
    label = _('Django file based backend')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()
        fields.update(
            {
                'file_path': {
                    'label': _('File path'),
                    'class': 'django.forms.CharField', 'kwargs': {
                        'max_length': 48
                    }
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Storage'), {
                    'fields': ('file_path',)
                }
            ),
        )

        return fieldsets



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'label', models.CharField(
                        unique=True, max_length=128, verbose_name='Label',
                        db_index=True
                    )
                ),
                (
                    'color', models.CharField(
                        max_length=3, verbose_name='Color', choices=[
                            ('blu', 'Blue'), ('cya', 'Cyan'), ('crl', 'Coral'),
                            ('gry', 'Green-Yellow'), ('kki', 'Khaki'),
                            ('lig', 'LightGrey'), ('mag', 'Magenta'),
                            ('red', 'Red'), ('org', 'Orange'),
                            ('yel', 'Yellow')
                        ]
                    )
                ),
                (
                    'documents', models.ManyToManyField(
                        related_name='tags', verbose_name='Documents',
                        to='documents.Document'
                    )
                ),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
            },
            bases=(models.Model,),
        ),
    ]



from mayan.apps.converter.transformations import (
    BaseTransformation, TransformationRotate90
)
from mayan.apps.converter.models import LayerTransformation
from mayan.apps.converter.tests.mixins import LayerTestMixin

from ..document_file_actions import DocumentFileActionAppendNewPages

from .base import GenericDocumentTestCase
from .mixins.document_file_mixins import DocumentFileTestMixin


class DocumentVersionPageTestCase(
    DocumentFileTestMixin, LayerTestMixin, GenericDocumentTestCase
):
    def _get_test_document_version_page_cached_image(self):
        cache_file = self._test_document_version_page.cache_partition.get_file(
            filename=self._test_document_version_page.get_combined_cache_filename()
        )
        with cache_file.open() as file_object:
            return file_object.read()

    def test_version_pages_reset_no_file(self):
        self._test_document_file.delete()
        self._test_document_version.pages_reset()

    def test_version_pages_reset(self):
        self._upload_test_document_file(
            action_name=DocumentFileActionAppendNewPages.backend_id
        )

        test_document_version_page_content_objects = self._test_document.versions.last().page_content_objects

        self._test_document_version.pages_reset()

        self.assertNotEqual(
            self._test_document.versions.last().page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self._test_document.versions.last().page_content_objects,
            list(self._test_document.file_latest.pages.all())
        )

    def test_method_get_absolute_url(self):
        self.assertTrue(
            self._test_document.version_active.pages.first().get_absolute_url()
        )

    def test_version_page_cache_update_on_transformation(self):
        BaseTransformation.register(
            layer=self._test_layer, transformation=TransformationRotate90
        )

        test_combined_cache_filename_1 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_1 = self._test_document_version_page.generate_image()
        test_api_image_url_1 = self._test_document_version_page.get_api_image_url()
        test_image_1 = self._get_test_document_version_page_cached_image()

        self._test_layer.add_transformation_to(
            obj=self._test_document_version_page,
            transformation_class=TransformationRotate90, arguments={}
        )

        test_combined_cache_filename_2 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_2 = self._test_document_version_page.generate_image()
        test_api_image_url_2 = self._test_document_version_page.get_api_image_url()
        test_image_2 = self._get_test_document_version_page_cached_image()

        self.assertNotEqual(test_combined_cache_filename_1, test_combined_cache_filename_2)
        self.assertNotEqual(test_generate_image_1, test_generate_image_2)
        self.assertNotEqual(test_api_image_url_1, test_api_image_url_2)
        self.assertNotEqual(test_image_1, test_image_2)

    def test_version_page_cache_update_on_source_transformation(self):
        BaseTransformation.register(
            layer=self._test_layer, transformation=TransformationRotate90
        )

        test_document_version_page_content_object = self._test_document_version_page.content_object

        # Original.
        test_combined_cache_filename_1 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_1 = self._test_document_version_page.generate_image()
        test_api_image_url_1 = self._test_document_version_page.get_api_image_url()
        # Source is at 0 deg, version is at 0 deg = image is at 0 deg.
        test_image_1 = self._get_test_document_version_page_cached_image()

        # Add transformation to content object.
        self._test_layer.add_transformation_to(
            obj=test_document_version_page_content_object,
            transformation_class=TransformationRotate90, arguments={}
        )

        test_combined_cache_filename_2 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_2 = self._test_document_version_page.generate_image()
        test_api_image_url_2 = self._test_document_version_page.get_api_image_url()
        # Source is at 90 deg, version is at 0 deg = image is at 90 deg.
        test_image_2 = self._get_test_document_version_page_cached_image()

        self.assertNotEqual(test_combined_cache_filename_1, test_combined_cache_filename_2)
        self.assertNotEqual(test_generate_image_1, test_generate_image_2)
        self.assertNotEqual(test_api_image_url_1, test_api_image_url_2)
        self.assertNotEqual(test_image_1, test_image_2)

        # Add transformation to document version page.
        self._test_layer.add_transformation_to(
            obj=self._test_document_version_page,
            transformation_class=TransformationRotate90, arguments={}
        )

        test_combined_cache_filename_3 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_3 = self._test_document_version_page.generate_image()
        test_api_image_url_3 = self._test_document_version_page.get_api_image_url()
        # Source is at 90 deg, version is at 90 deg = image is at 180 deg.
        test_image_3 = self._get_test_document_version_page_cached_image()

        self.assertNotEqual(test_combined_cache_filename_2, test_combined_cache_filename_3)
        self.assertNotEqual(test_generate_image_2, test_generate_image_3)
        self.assertNotEqual(test_api_image_url_2, test_api_image_url_3)
        self.assertNotEqual(test_image_2, test_image_3)

        self.assertNotEqual(test_combined_cache_filename_1, test_combined_cache_filename_3)
        self.assertNotEqual(test_generate_image_1, test_generate_image_3)
        self.assertNotEqual(test_api_image_url_1, test_api_image_url_3)
        self.assertNotEqual(test_image_1, test_image_3)

        # Remove transformation from document content object.
        LayerTransformation.objects.get_for_object(
            obj=test_document_version_page_content_object
        ).delete()

        test_combined_cache_filename_4 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_4 = self._test_document_version_page.generate_image()
        test_api_image_url_4 = self._test_document_version_page.get_api_image_url()
        # Source is at 0 deg, version is at 90 deg = image is at 90 deg.
        test_image_4 = self._get_test_document_version_page_cached_image()

        self.assertNotEqual(test_combined_cache_filename_1, test_combined_cache_filename_4)
        self.assertNotEqual(test_generate_image_1, test_generate_image_4)
        self.assertNotEqual(test_api_image_url_1, test_api_image_url_4)
        self.assertNotEqual(test_image_1, test_image_4)

        self.assertNotEqual(test_combined_cache_filename_2, test_combined_cache_filename_4)
        self.assertNotEqual(test_generate_image_2, test_generate_image_4)
        self.assertEqual(test_api_image_url_2, test_api_image_url_4)
        # Image 2 and 4 are equal

        self.assertNotEqual(test_combined_cache_filename_3, test_combined_cache_filename_4)
        self.assertNotEqual(test_generate_image_3, test_generate_image_4)
        self.assertNotEqual(test_api_image_url_3, test_api_image_url_4)
        self.assertNotEqual(test_image_3, test_image_4)

        # Remove transformation from document version page.
        LayerTransformation.objects.get_for_object(
            obj=self._test_document_version_page
        ).delete()

        test_combined_cache_filename_5 = self._test_document_version_page.get_combined_cache_filename()
        test_generate_image_5 = self._test_document_version_page.generate_image()
        test_api_image_url_5 = self._test_document_version_page.get_api_image_url()
        # Source is at 0 deg, version is at 0 deg = image is at 0 deg.
        test_image_5 = self._get_test_document_version_page_cached_image()

        self.assertEqual(test_combined_cache_filename_1, test_combined_cache_filename_5)
        self.assertEqual(test_generate_image_1, test_generate_image_5)
        self.assertEqual(test_api_image_url_1, test_api_image_url_5)
        self.assertEqual(test_image_1, test_image_5)

        self.assertNotEqual(test_combined_cache_filename_2, test_combined_cache_filename_5)
        self.assertNotEqual(test_generate_image_2, test_generate_image_5)
        self.assertNotEqual(test_api_image_url_2, test_api_image_url_5)
        self.assertNotEqual(test_image_2, test_image_5)

        self.assertNotEqual(test_combined_cache_filename_3, test_combined_cache_filename_5)
        self.assertNotEqual(test_generate_image_3, test_generate_image_5)
        self.assertNotEqual(test_api_image_url_3, test_api_image_url_5)
        self.assertNotEqual(test_image_3, test_image_5)

        self.assertNotEqual(test_combined_cache_filename_4, test_combined_cache_filename_5)
        self.assertNotEqual(test_generate_image_4, test_generate_image_5)
        self.assertNotEqual(test_api_image_url_4, test_api_image_url_5)
        self.assertNotEqual(test_image_4, test_image_5)

        self.assertEqual(test_combined_cache_filename_1, test_combined_cache_filename_5)
        self.assertEqual(test_generate_image_1, test_generate_image_5)
        self.assertEqual(test_api_image_url_1, test_api_image_url_5)
        self.assertEqual(test_image_1, test_image_5)



from django.apps import apps
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _


def get_groups():
    Group = apps.get_model(app_label='auth', model_name='Group')
    return ','.join(
        [
            group.name for group in Group.objects.all()
        ]
    )


def get_users():
    return ','.join(
        [
            user.get_full_name() or user.username
            for user in get_user_model().objects.all()
        ]
    )


def get_user_label_text(context):
    if not context['request'].user.is_authenticated:
        return _('Anonymous')
    else:
        return context['request'].user.get_full_name() or context['request'].user



import hashlib
import unittest

from fuse import FuseOSError

from django.db import connection
from django.test import tag

from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin

from ..filesystems import MirrorFilesystem
from ..runtime import cache

from .literals import (
    TEST_NODE_EXPRESSION, TEST_NODE_EXPRESSION_INVALID,
    TEST_NODE_EXPRESSION_MULTILINE, TEST_NODE_EXPRESSION_MULTILINE_EXPECTED,
    TEST_NODE_EXPRESSION_MULTILINE_2,
    TEST_NODE_EXPRESSION_MULTILINE_2_EXPECTED
)


@tag('mirroring')
@unittest.skipIf(
    condition=connection.vendor == 'mysql',
    reason='Known to fail due to unsupported feature of database manager.'
)
class IndexInstanceNodeMirroringTestCase(
    IndexTemplateTestMixin, GenericDocumentTestCase
):
    auto_create_test_index_template_node = False
    auto_upload_test_document = False

    def tearDown(self):
        cache.clear_all()
        super().tearDown()

    def _get_test_filesystem(self):
        def func_document_container_node():
            return self._test_index_template.index_template_root_node.get_index_instance_root_node()

        return MirrorFilesystem(
            func_document_container_node=func_document_container_node,
            node_text_attribute='value'
        )

    def test_document_access(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            test_filesystem.access(
                '/{}/{}'.format(
                    TEST_NODE_EXPRESSION, self._test_document.label
                )
            ), None
        )

    def test_trashed_document_access(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._create_test_document_stub()

        self._test_document.delete()

        test_filesystem = self._get_test_filesystem()
        with self.assertRaises(expected_exception=FuseOSError):
            self.assertEqual(
                test_filesystem.access(
                    '/{}/{}'.format(
                        TEST_NODE_EXPRESSION, self._test_document.label
                    )
                ), None
            )

    def test_document_access_failure(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )
        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()
        with self.assertRaises(expected_exception=FuseOSError):
            test_filesystem.access(
                '/{}/{}_non_valid'.format(
                    TEST_NODE_EXPRESSION, self._test_document.label
                )
            )

    def test_document_empty(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        test_filesystem = self._get_test_filesystem()

        self._upload_test_document()

        # Delete the physical document file without deleting the document
        # database entry.
        document_file = self._test_document.file_latest
        document_file.file.storage.delete(document_file.file.name)

        # Document file disk size is immutable since
        # 8733e07974c9ecce4880fb1fa19a14f6416c5fe6
        # 2022-04-30.
        self.assertEqual(
            test_filesystem.getattr(
                path='/{}/{}'.format(
                    TEST_NODE_EXPRESSION, self._test_document.label
                )
            )['st_size'], document_file.size
        )

    def test_document_open(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._upload_test_document()

        test_filesystem = self._get_test_filesystem()

        file_handle = test_filesystem.open(
            path='/{}/{}'.format(
                TEST_NODE_EXPRESSION, self._test_document.label
            ), flags='rb'
        )

        self.assertEqual(
            hashlib.sha256(
                test_filesystem.read(
                    fh=file_handle, offset=0, path=None,
                    size=self._test_document.file_latest.size
                )
            ).hexdigest(),
            self._test_document.file_latest.checksum
        )

    def test_multiline_indexes(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION_MULTILINE
        )

        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(test_filesystem.readdir('/', ''))[2:],
            [TEST_NODE_EXPRESSION_MULTILINE_EXPECTED]
        )

    def test_multiline_indexes_first_and_last(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION_MULTILINE_2
        )

        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(test_filesystem.readdir('/', ''))[2:],
            [TEST_NODE_EXPRESSION_MULTILINE_2_EXPECTED]
        )

    def test_stub_documents(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()

        self._test_index_template.rebuild()

        self.assertEqual(
            list(
                test_filesystem.readdir('/', '')
            )[2:], [TEST_NODE_EXPRESSION]
        )

    def test_duplicated_documents_readdir(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._create_test_document_stub()
        self._create_test_document_stub(label=self._test_document.label)

        test_filesystem = self._get_test_filesystem()

        self._test_index_template.rebuild()

        self.assertTrue(
            '{}({})'.format(
                self._test_documents[0].label, self._test_documents[0].pk
            ) in list(
                test_filesystem.readdir('/level_1', '')
            )
        )
        self.assertTrue(
            '{}({})'.format(
                self._test_documents[0].label, self._test_documents[1].pk
            ) in list(
                test_filesystem.readdir('/level_1', '')
            )
        )

    def test_duplicated_documents_open(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._upload_test_document()
        self._upload_test_document()

        test_filesystem = self._get_test_filesystem()

        self._test_index_template.rebuild()

        test_document_1_path = '/level_1/{}({})'.format(
            self._test_documents[0].label, self._test_documents[0].pk
        )
        test_document_2_path = '/level_1/{}({})'.format(
            self._test_documents[1].label, self._test_documents[1].pk
        )

        file_handle = test_filesystem.open(
            path=test_document_1_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_documents[0].file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

        file_handle = test_filesystem.open(
            path=test_document_2_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_documents[1].file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_document_label_character_open(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._upload_test_document()

        test_filesystem = self._get_test_filesystem()

        self._test_document.label = 'Document of 2019/12'
        self._test_document.save()

        self._test_index_template.rebuild()

        test_document_path = '/level_1/Document of 2019_12'

        file_handle = test_filesystem.open(
            path=test_document_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_document.file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_document_label_character_readdir(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION
        )

        self._upload_test_document()

        test_filesystem = self._get_test_filesystem()

        self._test_document.label = 'Document of 2019/12'
        self._test_document.save()

        self._test_index_template.rebuild()

        self.assertEqual(
            list(
                test_filesystem.readdir('/level_1', '')
            )[2], 'Document of 2019_12'
        )

    def test_invalid_directory_name_character_open(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION_INVALID
        )

        self._upload_test_document()

        test_filesystem = self._get_test_filesystem()

        self._test_index_template.rebuild()

        test_document_path = '/level_1/{}'.format(self._test_document.label)

        file_handle = test_filesystem.open(
            path=test_document_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_document.file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_directory_name_character_readdir(self):
        self._create_test_index_template_node(
            expression=TEST_NODE_EXPRESSION_INVALID
        )

        self._create_test_document_stub()

        test_filesystem = self._get_test_filesystem()

        self._test_index_template.rebuild()

        self.assertEqual(
            list(
                test_filesystem.readdir('/', '')
            )[2], 'level_1'
        )

        self.assertEqual(
            list(
                test_filesystem.readdir('/level_1', '')
            )[2], self._test_document.label
        )



from unittest import skip

from django.db import models

from mayan.apps.testing.tests.base import BaseTestCase

from ..search_backends import SearchBackend
from ..search_models import SearchModel

from .mixins.base import SearchTestMixin
from .mixins.task_mixins import SearchTaskTestMixin


class SearchTaskTestCase(SearchTaskTestMixin, SearchTestMixin, BaseTestCase):
    auto_create_test_object_model = True
    auto_create_test_object_fields = {
        'test_field': models.CharField(max_length=8)
    }
    auto_test_search_objects_create = False

    def _do_search(self, search_terms):
        return self._test_search_backend.search(
            search_model=self._test_model_search,
            query={
                'test_field': search_terms
            }, user=self._test_case_user
        )

    def _setup_test_model_search(self):
        self._test_model_search = SearchModel(
            app_label=self.TestModel._meta.app_label,
            model_name=self.TestModel._meta.model_name,
        )
        self._test_model_search.add_model_field(field='test_field')

    def setUp(self):
        super().setUp()
        self._create_test_object(
            instance_kwargs={'test_field': 'abc'}
        )

        backend = SearchBackend.get_instance()
        backend.reset()

    @skip(reason='Test with a backend that supports reindexing.')
    def test_task_index_instances(self):
        queryset = self._do_search(
            search_terms=self._test_objects[0].test_field
        )
        self.assertFalse(self._test_objects[0] in queryset)

        self._execute_task_index_instances()

        queryset = self._do_search(
            search_terms=self._test_objects[0].test_field
        )
        self.assertTrue(self._test_objects[0] in queryset)

    @skip(reason='Test with a backend that supports reindexing.')
    def test_task_reindex_backend(self):
        queryset = self._do_search(
            search_terms=self._test_objects[0].test_field
        )
        self.assertFalse(self._test_objects[0] in queryset)

        self._execute_task_reindex_backend()

        queryset = self._do_search(
            search_terms=self._test_objects[0].test_field
        )
        self.assertTrue(self._test_objects[0] in queryset)



from mayan.apps.appearance.classes import Icon
from mayan.apps.documents.icons import icon_document_type

icon_index = Icon(driver_name='fontawesome', symbol='list-ul')
icon_document_index_instance_list = Icon(
    driver_name='fontawesome', symbol='list-ul'
)
icon_document_type_index_templates = icon_index

icon_index_level_up = Icon(
    driver_name='fontawesomecss', css_classes='fa-level-up-alt fa-rotate-90'
)
icon_index_instance_list = Icon(driver_name='fontawesome', symbol='list-ul')
icon_index_instance_node_with_documents = Icon(
    driver_name='fontawesome', symbol='folder'
)
icon_index_instances_rebuild = Icon(
    driver_name='fontawesome-dual', primary_symbol='list-ul',
    secondary_symbol='hammer'
)
icon_index_instances_reset = Icon(
    driver_name='fontawesome-dual', primary_symbol='list-ul',
    secondary_symbol='times'
)

icon_index_template_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='list-ul',
    secondary_symbol='plus'
)
icon_index_template_delete = Icon(driver_name='fontawesome', symbol='times')
icon_index_template_document_types = icon_document_type
icon_index_template_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_index_template_event_triggers = Icon(
    driver_name='fontawesome', symbol='bolt'
)
icon_index_template_list = Icon(driver_name='fontawesome', symbol='list-ul')

icon_index_template_node_create = Icon(
    driver_name='fontawesome', symbol='plus'
)


icon_index_template_node_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_index_template_node_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)

icon_index_template_node_tree_view = Icon(
    driver_name='fontawesome', symbol='folder-open'
)



from furl import furl
from zipfile import ZipFile

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils.html import strip_tags
from django.utils.translation import ugettext_lazy as _

from mayan.apps.locales.utils import to_language
from mayan.apps.templating.classes import Template


class DocumentFileCompressor:
    context_key_list = (
        'document_list', 'download_file', 'download_list_url', 'download_url'
    )

    def __init__(self, queryset):
        self.queryset = queryset

    def compress(self, file_object, _event_action_object=None, _event_actor=None):
        with ZipFile(file=file_object, mode='w') as archive:
            for document_file in self.queryset.all():
                document_file._event_action_object = _event_action_object
                document_file._event_actor = _event_actor
                with document_file.get_download_file_object() as file_object:
                    archive.write(
                        arcname=str(document_file),
                        filename=file_object.name
                    )

    def compress_to_download_file(
        self, organization_installation_url='', filename=None, user=None
    ):
        # Hidden import
        from .settings import (
            setting_message_body_template, setting_message_subject_template
        )

        DownloadFile = apps.get_model(
            app_label='storage', model_name='DownloadFile'
        )
        Message = apps.get_model(
            app_label='messaging', model_name='Message'
        )

        if self.queryset.count():
            filename = filename or _('Document_file_bundle.zip')

            download_file = DownloadFile(
                filename=filename, label=_('Compressed document files'),
                user=user
            )
            download_file._event_actor = user
            download_file.save()

            with download_file.open(mode='wb+') as file_object:
                self.compress(
                    _event_action_object=download_file,
                    _event_actor=user, file_object=file_object
                )

            if user:
                download_list_url = furl(organization_installation_url).join(
                    reverse(
                        viewname='storage:download_file_list'
                    )
                ).tostr()

                download_url = furl(organization_installation_url).join(
                    reverse(
                        kwargs={'download_file_id': download_file.pk},
                        viewname='storage:download_file_download'
                    )
                ).tostr()

                context_dictionary = {
                    'document_list': self.queryset,
                    'download_file': download_file,
                    'download_list_url': download_list_url,
                    'download_url': download_url
                }

                if set(context_dictionary.keys()) != set(DocumentFileCompressor.context_key_list):
                    raise ImproperlyConfigured(
                        'The expected and the actual message template '
                        'context keys do not match. Update the expected key '
                        'list to match the actual key list being passed '
                        'to the template.'
                    )

                message_body_template = Template(
                    template_string=to_language(
                        language=user.locale_profile.language,
                        promise=setting_message_body_template.value
                    )
                )
                messge_body_content = message_body_template.render(
                    context=context_dictionary
                )

                message_subject_template = Template(
                    template_string=to_language(
                        language=user.locale_profile.language,
                        promise=setting_message_subject_template.value
                    )
                )
                message_subject_text = strip_tags(
                    message_subject_template.render(
                        context=context_dictionary
                    )
                )

                Message.objects.create(
                    body=messge_body_content, sender_object=download_file,
                    subject=message_subject_text, user=user
                )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0007_auto_20170803_0728'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='workflowtransition',
            name='trigger_time_period',
        ),
        migrations.RemoveField(
            model_name='workflowtransition',
            name='trigger_time_unit',
        ),
    ]



from django.db import models

from rest_framework.authtoken.models import Token

from mayan.apps.testing.tests.mixins import TestViewTestCaseMixin

from .. import serializers


class APIUserTestCaseMixin:
    def setUp(self):
        super().setUp()
        self._test_case_user_token = self.get_test_user_token(
            user=self._test_case_user
        )

    def get_test_user_token(self, user):
        token, created = Token.objects.get_or_create(
            user=user
        )

        return token


class DynamicFieldSerializerAPIViewTestCaseMixin:
    auto_add_test_view = True
    auto_create_test_object = False
    test_view_url = r'^test-view-url/(?P<test_object_id>\d+)/$'

    def _test_view_factory(self, test_object=None):
        self.TestModelParent = self._create_test_model(
            fields={
                'test_field_1': models.CharField(
                    blank=True, max_length=1
                ),
                'test_field_2': models.CharField(
                    blank=True, max_length=1
                )
            }, model_name='TestModelParent'
        )

        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                ),
                'test_field_3': models.CharField(
                    blank=True, max_length=1
                ),
                'test_field_4': models.CharField(
                    blank=True, max_length=1
                )
            }, model_name='TestModelChild'
        )

        self._test_object_parent = self.TestModelParent.objects.create()
        self._test_object_child = self.TestModelChild.objects.create(
            parent=self._test_object_parent
        )

        TestModelParent = self.TestModelParent
        TestModelChild = self.TestModelChild

        class TestModelParentSerializer(serializers.ModelSerializer):
            class Meta:
                fields = ('id', 'test_field_1', 'test_field_2')
                model = TestModelParent

        class TestModelChildSerializer(serializers.ModelSerializer):
            parent = TestModelParentSerializer()

            class Meta:
                fields = ('parent', 'id', 'test_field_3', 'test_field_4')
                model = TestModelChild

        class TestView(
            self._get_test_view_class(
                serializer_class=TestModelChildSerializer
            )
        ):
            """
            Flat subclass to allow the test class view to be called without
            code changes.
            """

        return TestView.as_view()

    def _request_test_api_view(self, query):
        return self.get(
            viewname='rest_api:{}'.format(self._test_view_name), kwargs={
                'test_object_id': self._test_object_child.pk,
            }, query=query
        )


class RESTAPIViewTestMixin:
    def _request_test_browser_api_view(self):
        return self.get(query={'format': 'api'}, viewname='rest_api:api_root')

    def _request_test_redoc_ui_view(self):
        return self.get(viewname='rest_api:schema-redoc')

    def _request_test_swagger_ui_view(self):
        return self.get(viewname='rest_api:schema-swagger-ui')

    def _request_test_swagger_no_ui_json_view(self):
        return self.get(
            kwargs={'format': '.json'}, viewname='rest_api:schema-json'
        )

    def _request_test_swagger_no_ui_yaml_view(self):
        return self.get(
            kwargs={'format': '.yaml'}, viewname='rest_api:schema-json'
        )


class TestAPIViewTestCaseMixin(TestViewTestCaseMixin):
    def _get_test_view_urlpatterns(self):
        from ..urls import api_version_urls
        return api_version_urls



from django.apps import apps
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_object, menu_related,
    menu_return, menu_secondary, menu_setup, menu_user
)
from mayan.apps.dashboards.dashboards import dashboard_administrator
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.logging.classes import ErrorLog
from mayan.apps.metadata.classes import MetadataLookup
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField
from mayan.apps.views.column_widgets import TwoStateWidget

from .dashboard_widgets import (
    DashboardWidgetGroupTotal, DashboardWidgetUserTotal
)
from .events import (
    event_group_created, event_group_edited, event_user_created,
    event_user_edited
)
from .handlers import handler_initialize_new_user_options
from .links import (
    link_current_user_details, link_group_create, link_group_edit,
    link_group_list, link_group_multiple_delete, link_group_user_list,
    link_group_setup, link_group_single_delete, link_user_create,
    link_user_edit, link_user_group_list, link_user_list,
    link_user_multiple_delete, link_user_set_options, link_user_setup,
    link_user_single_delete, separator_user_label, text_user_label
)
from .methods import (
    get_method_group_init, get_method_group_save, get_method_user_init,
    get_method_user_save, method_group_get_absolute_url,
    method_group_get_users, method_group_users_add,
    method_group_users_remove, method_user_get_absolute_api_url,
    method_user_get_absolute_url, method_user_get_groups,
    method_user_groups_add, method_user_groups_remove
)
from .permissions import (
    permission_group_delete, permission_group_edit,
    permission_group_view, permission_user_delete, permission_user_edit,
    permission_user_view
)
from .utils import get_groups, get_users


class UserManagementApp(MayanAppConfig):
    app_namespace = 'user_management'
    app_url = 'accounts'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.user_management'
    verbose_name = _('User management')

    def ready(self):
        super().ready()

        Group = apps.get_model(app_label='auth', model_name='Group')
        User = get_user_model()
        UserOptions = self.get_model(model_name='UserOptions')

        DynamicSerializerField.add_serializer(
            klass=get_user_model(),
            serializer_class='mayan.apps.user_management.serializers.UserSerializer'
        )

        Group._meta.ordering = ('name',)
        Group._meta.verbose_name = _('Group')
        Group._meta.verbose_name_plural = _('Groups')
        Group._meta.get_field(field_name='name').verbose_name = _('Name')

        Group.add_to_class(
            name='__init__', value=get_method_group_init()
        )
        Group.add_to_class(
            name='get_absolute_url', value=method_group_get_absolute_url
        )
        Group.add_to_class(
            name='get_users', value=method_group_get_users
        )
        Group.add_to_class(
            name='users_add', value=method_group_users_add
        )
        Group.add_to_class(
            name='users_remove', value=method_group_users_remove
        )
        Group.add_to_class(name='save', value=get_method_group_save())

        User._meta.ordering = ('pk',)
        User._meta.verbose_name = _('User')
        User._meta.verbose_name_plural = _('Users')
        User._meta.ordering = ('last_name', 'first_name')

        User._meta.get_field(field_name='email').verbose_name = _('Email')
        User._meta.get_field(
            field_name='first_name'
        ).verbose_name = _('First name')
        User._meta.get_field(field_name='groups').verbose_name = _('Groups')
        User._meta.get_field(
            field_name='is_active'
        ).verbose_name = _('Is active?')
        User._meta.get_field(
            field_name='is_superuser'
        ).verbose_name = _('Is super user?')
        User._meta.get_field(
            field_name='last_name'
        ).verbose_name = _('Last name')
        User._meta.get_field(
            field_name='password'
        ).verbose_name = _('Password')
        User._meta.get_field(
            field_name='username'
        ).verbose_name = _('Username')
        User._meta.get_field(
            field_name='last_login'
        ).verbose_name = _('Last login')

        User.add_to_class(
            name='__init__', value=get_method_user_init()
        )
        User.add_to_class(
            name='get_absolute_api_url',
            value=method_user_get_absolute_api_url
        )
        User.add_to_class(
            name='get_absolute_url', value=method_user_get_absolute_url
        )
        User.add_to_class(
            name='get_groups', value=method_user_get_groups
        )
        User.add_to_class(
            name='groups_add', value=method_user_groups_add
        )
        User.add_to_class(
            name='groups_remove', value=method_user_groups_remove
        )
        User.add_to_class(
            name='save', value=get_method_user_save()
        )

        User.has_usable_password.short_description = _(
            'Has usable password?'
        )

        EventModelRegistry.register(model=Group)
        EventModelRegistry.register(model=User)

        MetadataLookup(
            description=_('All the groups.'), name='groups',
            value=get_groups
        )
        MetadataLookup(
            description=_('All the users.'), name='users',
            value=get_users
        )

        ModelCopy(
            model=Group, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'name', 'user',
            )
        )
        ModelCopy(model=UserOptions).add_fields(
            field_names=('user', 'block_password_change'),
            field_value_templates={
                'id': '{user.user_options.id}'
            }
        )

        def condition_copy_user_model(instance):
            return not instance.is_superuser and not instance.is_staff

        ModelCopy(
            model=User, condition=condition_copy_user_model, bind_link=True,
            register_permission=True
        ).add_fields(
            field_names=(
                'username', 'first_name', 'last_name', 'email', 'is_active',
                'password', 'groups', 'user_options'
            )
        )

        ModelEventType.register(
            event_types=(event_group_created, event_group_edited),
            model=Group
        )

        ModelEventType.register(
            event_types=(event_user_created, event_user_edited), model=User
        )

        ModelPermission.register(
            model=Group, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_group_delete, permission_group_edit,
                permission_group_view
            )
        )
        ModelPermission.register(
            model=User, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_user_delete, permission_user_edit,
                permission_user_view
            )
        )

        SourceColumn(
            attribute='name', is_identifier=True, is_sortable=True,
            source=Group
        )
        SourceColumn(
            attribute='user_set.count', include_label=True, label=_('Users'),
            source=Group
        )

        SourceColumn(
            attribute='username', is_object_absolute_url=True,
            is_identifier=True, is_sortable=True, source=User
        )
        SourceColumn(
            attribute='first_name', include_label=True, is_sortable=True,
            source=User
        )
        SourceColumn(
            attribute='last_name', include_label=True, is_sortable=True,
            source=User
        )
        SourceColumn(
            attribute='email', include_label=True, is_sortable=True,
            source=User
        )
        SourceColumn(
            attribute='last_login', include_label=True, is_sortable=True,
            source=User
        )
        SourceColumn(
            attribute='is_active', include_label=True, is_sortable=True,
            source=User, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='is_superuser', include_label=True, is_sortable=True,
            source=User, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='has_usable_password', include_label=True, source=User,
            widget=TwoStateWidget
        )

        dashboard_administrator.add_widget(
            widget=DashboardWidgetUserTotal, order=99
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetGroupTotal, order=99
        )

        error_log = ErrorLog(app_config=self)
        error_log.register_model(model=User)

        # Group

        menu_list_facet.bind_links(
            links=(
                link_group_user_list,
            ), sources=(Group,)
        )
        menu_multi_item.bind_links(
            links=(link_group_multiple_delete,),
            sources=('user_management:group_list',)
        )
        menu_object.bind_links(
            links=(link_group_edit,),
            sources=(Group,)
        )
        menu_object.bind_links(
            links=(link_group_single_delete,), position=99,
            sources=(Group,)
        )
        menu_related.bind_links(
            links=(link_user_setup,), sources=(
                Group, 'user_management:group_multiple_delete',
                'user_management:group_list', 'user_management:group_create'

            )
        )
        menu_return.bind_links(
            links=(link_group_list,), sources=(
                Group, 'user_management:group_multiple_delete',
                'user_management:group_list', 'user_management:group_create'

            )
        )
        menu_secondary.bind_links(
            links=(link_group_create,), sources=(
                Group, 'user_management:group_multiple_delete',
                'user_management:group_list', 'user_management:group_create'
            )
        )

        # User

        menu_list_facet.bind_links(
            links=(
                link_user_group_list, link_user_set_options
            ), sources=(User,)
        )
        menu_multi_item.bind_links(
            links=(link_user_multiple_delete,),
            sources=('user_management:user_list',)
        )
        menu_object.bind_links(
            links=(
                link_user_single_delete, link_user_edit
            ), sources=(User,)
        )
        menu_related.bind_links(
            links=(link_group_setup,), sources=(
                User, 'authentication:user_multiple_set_password',
                'user_management:user_multiple_delete',
                'user_management:user_list', 'user_management:user_create'
            )
        )
        menu_return.bind_links(
            links=(link_user_list,), sources=(
                User, 'authentication:user_multiple_set_password',
                'user_management:user_multiple_delete',
                'user_management:user_list', 'user_management:user_create'
            )
        )
        menu_secondary.bind_links(
            links=(link_user_create,), sources=(
                User, 'authentication:user_multiple_set_password',
                'user_management:user_multiple_delete',
                'user_management:user_list', 'user_management:user_create'
            )
        )
        menu_setup.bind_links(
            links=(link_user_setup, link_group_setup)
        )
        menu_user.bind_links(
            links=(
                text_user_label, separator_user_label,
                link_current_user_details
            ), position=0
        )

        post_save.connect(
            dispatch_uid='user_management_handler_initialize_new_user_options',
            receiver=handler_initialize_new_user_options,
            sender=User
        )



from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.literals import (
    TIME_DELTA_UNIT_CHOICES, TIME_DELTA_UNIT_DAYS
)
from mayan.apps.databases.model_mixins import (
    ExtraDataModelMixin, ModelMixinConditionField
)
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import event_workflow_template_edited

from .workflow_state_escalation_model_mixins import WorkflowStateEscalationBusinessLogicMixin
from .workflow_state_models import WorkflowState
from .workflow_transition_models import WorkflowTransition

__all__ = ('WorkflowStateEscalation',)


class WorkflowStateEscalation(
    ExtraDataModelMixin, ModelMixinConditionField,
    WorkflowStateEscalationBusinessLogicMixin, models.Model
):
    state = models.ForeignKey(
        on_delete=models.CASCADE, related_name='escalations',
        to=WorkflowState, verbose_name=_('Workflow state')
    )
    transition = models.ForeignKey(
        on_delete=models.CASCADE, related_name='escalations',
        to=WorkflowTransition, verbose_name=_('Transition')
    )
    priority = models.IntegerField(
        blank=True, db_index=True, default=0, help_text=_(
            'Determine the order in which the escalations will be '
            'evaluated. Escalations with a lower priority number are '
            'executed before escalations with a higher priority number.'
        ), verbose_name=_('Priority')
    )
    enabled = models.BooleanField(
        default=True, help_text=_(
            'Enable automatic transition the workflow after a specified '
            'amount of time has elapsed in the state without change.'
        ), verbose_name=_('Enabled')
    )
    unit = models.CharField(
        blank=True, choices=TIME_DELTA_UNIT_CHOICES,
        default=TIME_DELTA_UNIT_DAYS, max_length=32, verbose_name=_(
            'Time unit'
        )
    )
    amount = models.PositiveIntegerField(
        blank=True, default=1, help_text=_(
            'Amount of the selected escalation units of time.'
        ), verbose_name=_('Amount')
    )
    comment = models.TextField(
        blank=True, help_text=_(
            'Comment to save to the workflow instance when the escalation '
            'is executed.'
        ), verbose_name=_('Comment')
    )

    class Meta:
        ordering = ('priority',)
        unique_together = ('state', 'transition')
        verbose_name = _('Workflow state escalation')
        verbose_name_plural = _('Workflow state escalations')

    def __str__(self):
        return '{} -> {} ({})'.format(
            self.state, self.transition, self.priority
        )

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='state.workflow'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'state.workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'state.workflow'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



import json

from mayan.apps.document_states.events import event_workflow_instance_transitioned
from mayan.apps.document_states.literals import WORKFLOW_ACTION_ON_ENTRY
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from mayan.apps.document_states.tests.mixins.workflow_template_transition_mixins import WorkflowTemplateTransitionTestMixin
from mayan.apps.documents.tests.base import (
    GenericDocumentTestCase, GenericDocumentViewTestCase
)

from ..events import event_message_created
from ..models import Message
from ..workflow_actions import WorkflowActionMessageSend

from .literals import TEST_MESSAGE_BODY, TEST_MESSAGE_SUBJECT


class WorkflowActionMessageSendTestCase(
    WorkflowTemplateStateActionTestMixin, GenericDocumentTestCase
):
    auto_create_test_message = False

    def test_message_send_workflow_action(self):
        test_message_count = Message.objects.count()

        self._clear_events()

        self._execute_workflow_template_state_action(
            klass=WorkflowActionMessageSend, kwargs={
                'body': TEST_MESSAGE_BODY,
                'subject': TEST_MESSAGE_SUBJECT,
                'username_list': self._test_case_user.username
            }
        )
        self.assertEqual(
            Message.objects.count(), test_message_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        _test_message = Message.objects.first()

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, _test_message)
        self.assertEqual(events[0].target, _test_message)
        self.assertEqual(events[0].verb, event_message_created.id)


class WorkflowActionMessageSendViewTestCase(
    WorkflowTemplateStateActionViewTestMixin,
    WorkflowTemplateTransitionTestMixin, GenericDocumentViewTestCase
):
    auto_create_test_message = False
    auto_create_test_workflow_template = False
    auto_create_test_workflow_template_state = False
    auto_upload_test_document = False

    def test_message_send_workflow_action(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

        backend_data = json.dumps(
            obj={
                'body': TEST_MESSAGE_BODY,
                'subject': TEST_MESSAGE_SUBJECT,
                'username_list': self._test_case_user.username
            }
        )

        self._test_workflow_template_states[1].actions.create(
            backend_data=backend_data,
            backend_path=WorkflowActionMessageSend.backend_id,
            label='', when=WORKFLOW_ACTION_ON_ENTRY,
        )
        self._test_workflow_template.document_types.add(
            self._test_document_type
        )

        self._create_test_document_stub()

        self._clear_events()

        self._test_document.workflows.first().do_transition(
            transition=self._test_workflow_template_transition
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        _test_message = Message.objects.first()

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, _test_message)
        self.assertEqual(events[0].target, _test_message)
        self.assertEqual(events[0].verb, event_message_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(
            events[1].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[1].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[1].verb, event_workflow_instance_transitioned.id
        )



from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Message

from .literals import TEST_MESSAGE_BODY, TEST_MESSAGE_SUBJECT


class MessageTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Message
    _test_object_name = '_test_message'
    auto_create_test_message = True

    def setUp(self):
        super().setUp()

        if self.auto_create_test_message:
            self._create_test_message()

    def _create_test_message(self):
        self._test_message = Message.objects.create(
            body=TEST_MESSAGE_BODY, subject=TEST_MESSAGE_SUBJECT,
            user=self._test_case_user
        )


class MessageAPIViewTestMixin(MessageTestMixin):
    def _request_test_message_create_api_view(self, extra_data=None):
        self._test_object_track()

        data = {
            'body': TEST_MESSAGE_BODY,
            'subject': TEST_MESSAGE_SUBJECT,
            'user': self._test_case_user.pk
        }

        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='rest_api:message-list', data=data
        )

        self._test_object_set()

        return response

    def _request_test_message_delete_api_view(self):
        return self.delete(
            viewname='rest_api:message-detail',
            kwargs={'message_id': self._test_message.pk}
        )

    def _request_test_message_detail_api_view(self):
        return self.get(
            viewname='rest_api:message-detail',
            kwargs={'message_id': self._test_message.pk}
        )

    def _request_test_message_edit_api_view(
        self, extra_data=None, verb='patch'
    ):
        data = {
            'read': True,
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, verb)(
            viewname='rest_api:message-detail',
            kwargs={'message_id': self._test_message.pk},
            data=data
        )

    def _request_test_message_list_api_view(self):
        return self.get(viewname='rest_api:message-list')


class MessageViewTestMixin(MessageTestMixin):
    def _request_test_message_create_view(self, extra_data=None):
        self._test_object_track()

        data = {
            'body': TEST_MESSAGE_BODY,
            'subject': TEST_MESSAGE_SUBJECT,
            'user': self._test_case_user.pk
        }

        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='messaging:message_create', data=data
        )

        self._test_object_set()

        return response

    def _request_test_message_delete_view(self):
        return self.post(
            viewname='messaging:message_single_delete', kwargs={
                'message_id': self._test_message.pk
            }
        )

    def _request_test_message_detail_view(self):
        return self.get(
            viewname='messaging:message_detail', kwargs={
                'message_id': self._test_message.pk
            }
        )

    def _request_test_message_list_view(self):
        return self.get(viewname='messaging:message_list')

    def _request_test_message_mark_all_read_view(self):
        return self.post(viewname='messaging:message_all_mark_read')

    def _request_test_message_mark_read_view(self):
        return self.post(
            viewname='messaging:message_single_mark_read', kwargs={
                'message_id': self._test_message.pk
            }
        )

    def _request_test_message_mark_unread_view(self):
        return self.post(
            viewname='messaging:message_single_mark_unread', kwargs={
                'message_id': self._test_message.pk
            }
        )



from unittest import mock

import mayan

from mayan.apps.testing.tests.base import BaseTestCase

from ..utils import PyPIClient


class PyPIClientTestCase(BaseTestCase):
    def test_method_get_server_proxy(self):
        PyPIClient().get_server_proxy()

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_not_latest_version(self, mock_package_releases):
        mock_package_releases.return_value = ('99.99.99',)
        with self.assertRaises(expected_exception=PyPIClient.ExceptionNotLatestVersion):
            PyPIClient().check_version()

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_unknown_version(self, mock_package_releases):
        mock_package_releases.return_value = None
        with self.assertRaises(expected_exception=PyPIClient.ExceptionUnknownLatestVersion):
            PyPIClient().check_version()

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_correct_version(self, mock_package_releases):
        mock_package_releases.return_value = (mayan.__version__,)
        PyPIClient().check_version()

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_ahead_of_upstream(self, mock_package_releases):
        mock_package_releases.return_value = ('1.0.0',)
        with self.assertRaises(expected_exception=PyPIClient.ExceptionAheadOfUpstream):
            PyPIClient().check_version()



from rest_framework import status

from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase
from mayan.apps.sources.events import event_source_created
from mayan.apps.sources.models import Source
from mayan.apps.sources.permissions import (
    permission_sources_create, permission_sources_edit
)
from mayan.apps.sources.tests.mixins.source_api_view_mixins import (
    SourceAPIViewTestMixin, SourceActionAPIViewTestMixin
)

from .mixins import WatchStorageSourceTestMixin


class WatchStorageSourceBackendAPIViewTestCase(
    DocumentTestMixin, SourceAPIViewTestMixin, WatchStorageSourceTestMixin,
    BaseAPITestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = True

    def test_source_create_api_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_created.id)


class WatchStorageSourceBackendActionFileDeleteAPIViewTestCase(
    SourceActionAPIViewTestMixin, WatchStorageSourceTestMixin,
    DocumentTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def test_basic_no_permission(self):
        self.copy_test_source_file()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self.copy_test_source_file()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WatchStorageSourceBackendActionDocumentUploadAPIViewTestCase(
    SourceActionAPIViewTestMixin, WatchStorageSourceTestMixin,
    DocumentTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def test_basic_no_permission(self):
        self.copy_test_source_file()

        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view(
            action_name='document_upload'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), test_document_count)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        self.copy_test_source_file()

        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view(
            action_name='document_upload'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(Document.objects.count(), test_document_count)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.apps import apps
from django.utils.translation import ugettext_lazy as _


def method_document_get_tags(self, permission, user):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    DocumentTag = apps.get_model(app_label='tags', model_name='DocumentTag')

    return AccessControlList.objects.restrict_queryset(
        permission=permission, queryset=DocumentTag.objects.filter(
            documents=self
        ), user=user
    )


method_document_get_tags.help_text = _(
    'Return the tags attached to the document.'
)
method_document_get_tags.short_description = _('get_tags()')



from shutil import copyfileobj

import sh

from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.exceptions import DependenciesException
from mayan.apps.storage.utils import NamedTemporaryFile

from ..classes import MIMETypeBackend

from .literals import DEFAULT_FILE_PATH


class MIMETypeBackendFileCommand(MIMETypeBackend):
    def _init(self, copy_length=None, file_path=None):
        self.file_path = file_path or DEFAULT_FILE_PATH
        self.copy_length = copy_length

        try:
            self.command_file = sh.Command(path=self.file_path).bake(
                brief=True, mime_type=True
            )
        except sh.CommandNotFound:
            raise DependenciesException(
                _('file command not installed or not found.')
            )

    def _get_mime_type(self, file_object, mime_type_only):
        with NamedTemporaryFile() as temporary_file_object:
            file_object.seek(0)
            copyfileobj(
                fsrc=file_object, fdst=temporary_file_object,
                length=self.copy_length
            )
            file_object.seek(0)
            temporary_file_object.seek(0)

            output = self.command_file(
                temporary_file_object.name, mime_encoding=not mime_type_only
            ).split(';')

            file_mime_type = output[0]

            if mime_type_only:
                file_mime_encoding = 'binary'
            else:
                # Remove the ' charset=' string from the output.
                file_mime_encoding = output[1].split(' charset=')[1]

            # Remove trailing newline.
            file_mime_type = file_mime_type.split('\n')[0]
            file_mime_encoding = file_mime_encoding.split('\n')[0]

            return (file_mime_type, file_mime_encoding)



from django.conf.urls import url

from .api_views import (
    APIDocumentTypeOCRSettingsView, APIDocumentOCRSubmitView,
    APIDocumentVersionOCRSubmitView,
    APIDocumentVersionPageOCRContentDetailView
)
from .views import (
    DocumentVersionOCRContentDeleteView, DocumentVersionOCRContentView,
    DocumentVersionOCRDownloadView, DocumentVersionPageOCRContentEditView,
    DocumentVersionPageOCRContentDetailView, DocumentVersionOCRSubmitView,
    DocumentTypeOCRSettingsEditView, DocumentTypeOCRSubmitView
)

urlpatterns_document_types = [
    url(
        regex=r'^document_types/submit/$', name='document_type_submit',
        view=DocumentTypeOCRSubmitView.as_view()
    ),
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/settings/$',
        name='document_type_ocr_settings',
        view=DocumentTypeOCRSettingsEditView.as_view()
    )
]

urlpatterns_document_versions = [
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/content/$',
        name='document_version_ocr_content_view',
        view=DocumentVersionOCRContentView.as_view()
    ),
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/content/delete/$',
        name='document_version_ocr_content_single_delete',
        view=DocumentVersionOCRContentDeleteView.as_view()
    ),
    url(
        regex=r'^documents/versions/multiple/content/delete/$',
        name='document_version_ocr_content_multiple_delete',
        view=DocumentVersionOCRContentDeleteView.as_view()
    ),
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/download/$',
        name='document_version_ocr_content_download',
        view=DocumentVersionOCRDownloadView.as_view()
    ),
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/submit/$',
        name='document_version_ocr_single_submit',
        view=DocumentVersionOCRSubmitView.as_view()
    ),
    url(
        regex=r'^documents/versions/multiple/submit/$',
        name='document_version_ocr_multiple_submit',
        view=DocumentVersionOCRSubmitView.as_view()
    ),
    url(
        regex=r'^documents/versions/pages/(?P<document_version_page_id>\d+)/content/edit/$',
        name='document_version_page_ocr_content_edit_view',
        view=DocumentVersionPageOCRContentEditView.as_view()
    ),
    url(
        regex=r'^documents/versions/pages/(?P<document_version_page_id>\d+)/content/$',
        name='document_version_page_ocr_content_detail_view',
        view=DocumentVersionPageOCRContentDetailView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document_types)
urlpatterns.extend(urlpatterns_document_versions)

api_urls = [
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/ocr/settings/$',
        name='document-type-ocr-settings-view',
        view=APIDocumentTypeOCRSettingsView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/ocr/submit/$',
        name='document-ocr-submit-view',
        view=APIDocumentOCRSubmitView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/versions/(?P<document_version_id>\d+)/ocr/submit/$',
        name='document-version-ocr-submit-view',
        view=APIDocumentVersionOCRSubmitView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/versions/(?P<document_version_id>\d+)/pages/(?P<document_version_page_id>\d+)/ocr/$',
        name='document-version-page-ocr-content-detail-view',
        view=APIDocumentVersionPageOCRContentDetailView.as_view()
    )
]



import os

from django.conf import settings

TEST_DEPENDENCY_GROUP_NAME = 'environment'
TEST_DEPENDENCY_GROUP_ENTRY_NAME = 'testing'

TEST_TAR_CVE_2007_4559_FILENAME = 'CVE_2007_4559.tar'

TEST_TAR_CVE_2007_4559_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'dependencies', 'tests', 'contrib',
    'test_files', TEST_TAR_CVE_2007_4559_FILENAME
)


MESSAGE_TEST_MORE_RECENT = 'more recent'
MESSAGE_TEST_NOT_LATEST = 'is outdated'
MESSAGE_TEST_UNKNOWN_VERSION = 'not possible'
MESSAGE_TEST_UP_TO_DATE = 'up-to-date'



import logging

from django.apps import apps
from django.contrib.auth import get_user_model

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_document_version_page_list_append(
    document_version_id, user_id=None
):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )
    document_version.pages_append_all(user=user)


@app.task(ignore_result=True)
def task_document_version_page_list_reset(document_version_id, user_id=None):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )
    document_version.pages_reset(user=user)


@app.task(ignore_result=True)
def task_document_version_delete(document_version_id, user_id=None):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    document_version._event_actor = user
    document_version.delete()


@app.task(ignore_result=True)
def task_document_version_export(
    document_version_id, organization_installation_url=None, user_id=None
):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    document_version.export_to_download_file(
        organization_installation_url=organization_installation_url,
        user=user
    )



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='Whoosh', version_string='==2.7.4'
)
PythonDependency(
    module=__name__, name='dateparser', version_string='==1.1.8'
)
PythonDependency(
    module=__name__, name='elasticsearch', version_string='==7.17.9'
)
PythonDependency(
    module=__name__, name='elasticsearch-dsl', version_string='==7.4.1'
)



from django.conf.urls import url

from .api_views import (
    APIDocumentFilePageContentView, APIDocumentTypeParsingSettingsView
)
from .views import (
    DocumentFileContentView, DocumentFileContentDeleteView,
    DocumentFileContentDownloadView, DocumentFilePageContentView,
    DocumentFileSubmitView, DocumentTypeSettingsEditView,
    DocumentTypeSubmitView
)

urlpatterns_document_files = [
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/content/$',
        name='document_file_content_view',
        view=DocumentFileContentView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/content/delete/$',
        name='document_file_content_single_delete',
        view=DocumentFileContentDeleteView.as_view()
    ),
    url(
        regex=r'^documents/files/multiple/content/delete/$',
        name='document_file_content_multiple_delete',
        view=DocumentFileContentDeleteView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/content/download/$',
        name='document_file_content_download',
        view=DocumentFileContentDownloadView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/submit/$',
        name='document_file_parsing_single_submit',
        view=DocumentFileSubmitView.as_view()
    ),
    url(
        regex=r'^documents/files/multiple/submit/$',
        name='document_file_parsing_multiple_submit',
        view=DocumentFileSubmitView.as_view()
    ),
    url(
        regex=r'^documents/files/pages/(?P<document_file_page_id>\d+)/content/$',
        name='document_file_page_content_view',
        view=DocumentFilePageContentView.as_view()
    )
]

urlpatterns_document_types = [
    url(
        regex=r'^document_types/submit/$', name='document_type_submit',
        view=DocumentTypeSubmitView.as_view()
    ),
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/parsing/settings/$',
        name='document_type_parsing_settings',
        view=DocumentTypeSettingsEditView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document_files)
urlpatterns.extend(urlpatterns_document_types)

api_urls = [
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/parsing/settings/$',
        name='document-type-parsing-settings-view',
        view=APIDocumentTypeParsingSettingsView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/files/(?P<document_file_id>\d+)/pages/(?P<document_file_page_id>\d+)/content/$',
        name='document-file-page-content-view',
        view=APIDocumentFilePageContentView.as_view()
    )
]



import os

from django.conf import settings
from django.db.models import Q

from mayan.apps.converter.classes import Layer

from ...literals import PAGE_RANGE_ALL
from ...models import Document

from ..literals import (
    DEFAULT_DOCUMENT_STUB_LABEL, TEST_DOCUMENT_DESCRIPTION,
    TEST_DOCUMENT_DESCRIPTION_EDITED, TEST_DOCUMENT_LABEL,
    TEST_FILE_SMALL_FILENAME, TEST_FILE_SMALL_PATH
)

from .document_type_mixins import DocumentTypeTestMixin


class DocumentAPIViewTestMixin:
    def _request_test_document_change_type_api_view(self):
        return self.post(
            viewname='rest_api:document-change-type', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'document_type_id': self._test_document_types[1].pk
            }
        )

    def _request_test_document_create_api_view(self):
        pk_list = list(
            Document.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:document-list', data={
                'document_type_id': self._test_document_type.pk
            }
        )

        try:
            self._test_document = Document.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Document.DoesNotExist:
            self._test_document = None

        return response

    def _request_test_document_detail_api_view(self):
        return self.get(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self._test_document.pk
            }, data={'description': TEST_DOCUMENT_DESCRIPTION_EDITED}
        )

    def _request_test_document_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self._test_document.pk
            }, data={'description': TEST_DOCUMENT_DESCRIPTION_EDITED}
        )

    def _request_test_document_list_api_view(self):
        return self.get(viewname='rest_api:document-list')

    def _request_test_document_upload_api_view(self):
        pk_list = list(
            Document.objects.values_list('pk', flat=True)
        )

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            response = self.post(
                viewname='rest_api:document-upload', data={
                    'document_type_id': self._test_document_type.pk,
                    'file': file_object
                }
            )

        try:
            self._test_document = Document.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Document.DoesNotExist:
            self._test_document = None

        return response


class DocumentTestMixin(DocumentTypeTestMixin):
    _test_document_count = 1
    _test_document_file_filename = TEST_FILE_SMALL_FILENAME
    _test_document_file_path = None
    _test_document_filename = TEST_FILE_SMALL_FILENAME
    _test_document_language = None
    _test_document_path = None
    auto_create_test_document_stub = False
    auto_upload_test_document = True

    def setUp(self):
        super().setUp()
        Layer.invalidate_cache()

        self._test_documents = []
        self._test_document_file_list = []
        self._test_document_file_page_list = []
        self._test_document_id_list = []
        self._test_document_id_list_string = []
        self._test_document_version_list = []
        self._test_document_version_page_list = []

        if self.auto_create_test_document_type:
            if self._test_document_count > 1:
                if self.auto_upload_test_document:
                    self._upload_test_documents()
                elif self.auto_create_test_document_stub:
                    self._create_test_document_stubs()
            else:
                if self.auto_upload_test_document:
                    self._upload_test_document()
                elif self.auto_create_test_document_stub:
                    self._create_test_document_stub()

    def _calculate_test_document_file_path(self):
        if not self._test_document_file_path:
            self._test_document_file_path = os.path.join(
                settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
                'sample_documents', self._test_document_file_filename
            )

    def _calculate_test_document_path(self):
        if not self._test_document_path:
            self._test_document_path = os.path.join(
                settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
                'sample_documents', self._test_document_filename
            )

    def _create_test_document_stub(self, document_type=None, label=None):
        if label is None:
            label = '{}_{}'.format(
                DEFAULT_DOCUMENT_STUB_LABEL, len(self._test_documents)
            )

        self._test_document_stub = Document.objects.create(
            document_type=document_type or self._test_document_type,
            label=label
        )
        self._test_document = self._test_document_stub
        self._test_documents.append(self._test_document)
        self._test_document_id_list.append(self._test_document.pk)
        self._test_document_id_list_string.append(
            str(self._test_document.pk)
        )

    def _create_test_document_stubs(self, count=None):
        for index in range(count or self._test_document_count):
            self._create_test_document_stub()

    def _create_test_documents(self, count=None):
        for index in range(count or self._test_document_count):
            self._upload_test_document(
                label='{}_{}'.format(
                    TEST_DOCUMENT_LABEL,
                    len(self._test_documents)
                )
            )

    def _upload_test_document(
        self, description=None, document_file_attributes=None,
        document_type=None, document_version_attributes=None, label=None,
        user=None
    ):
        self._calculate_test_document_path()

        if not label:
            label = self._test_document_filename

        test_document_description = description or '{}_{}'.format(
            TEST_DOCUMENT_DESCRIPTION, len(self._test_documents)
        )

        document_type = document_type or self._test_document_type

        with open(file=self._test_document_path, mode='rb') as file_object:
            document = document_type.documents_upload(
                description=test_document_description,
                file_object=file_object, label=label,
                language=self._test_document_language, user=user
            )

        self._test_document = document
        self._test_documents.append(self._test_document)
        self._test_document_id_list.append(self._test_document.pk)
        self._test_document_id_list_string.append(
            str(self._test_document.pk)
        )

        self._test_document_file = document.file_latest
        self._test_document_file_list.append(self._test_document_file)
        self._test_document_file_page_list = list(
            self._test_document_file.file_pages.all()
        )
        self._test_document_file_page = self._test_document_file.file_pages.first()
        self._test_document_version = self._test_document.version_active
        self._test_document_version_list.append(self._test_document_version)
        self._test_document_version_page_list = list(
            self._test_document_version.version_pages.all()
        )
        self._test_document_version_page = self._test_document_version.version_pages.first()

        if document_file_attributes:
            for key, value in document_file_attributes.items():
                setattr(self._test_document_file, key, value)

            self._test_document_file.save()

        if document_version_attributes:
            for key, value in document_version_attributes.items():
                setattr(self._test_document_version, key, value)

            self._test_document_version.save()


class DocumentViewTestMixin:
    def _request_test_document_list_view(self, data=None):
        return self.get(viewname='documents:document_list', data=data)

    def _request_test_document_preview_view(self):
        return self.get(
            viewname='documents:document_preview', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_print_form_view(self):
        return self.get(
            viewname='documents:document_print_form', kwargs={
                'document_id': self._test_document.pk,
            }, data={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_print_view(self):
        return self.get(
            viewname='documents:document_print_view', kwargs={
                'document_id': self._test_document.pk,
            }, query={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_properties_edit_get_view(self):
        return self.get(
            viewname='documents:document_properties_edit', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_properties_view(self):
        return self.get(
            viewname='documents:document_properties', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_type_change_get_view(self):
        return self.get(
            viewname='documents:document_type_change', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_type_change_post_view(self):
        return self.post(
            viewname='documents:document_type_change', kwargs={
                'document_id': self._test_document.pk
            }, data={'document_type': self._test_document_types[1].pk}
        )

    def _request_test_document_multiple_type_change(self):
        return self.post(
            viewname='documents:document_multiple_type_change',
            data={
                'id_list': self._test_document.pk,
                'document_type': self._test_document_types[1].pk
            }
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.links import link_transformation_list

from .layers import layer_redactions

link_redaction_list = link_transformation_list.copy(
    layer=layer_redactions
)
link_redaction_list.text = _('Redactions')



from mayan.apps.authentication.forms import AuthenticationFormBase


class AuthenticationFormOIDC(AuthenticationFormBase):
    def __init__(self, **kwargs):
        kwargs.update(
            {
                'data': None, 'files': None
            }
        )
        super().__init__(**kwargs)



import logging

import gnupg

from mayan.apps.storage.utils import TemporaryDirectory

from ..classes import GPGBackend
from ..literals import DEFAULT_GPG_PATH
from ..settings import setting_gpg_backend_arguments

gpg_path = setting_gpg_backend_arguments.value.get(
    'gpg_path', DEFAULT_GPG_PATH
)
logger = logging.getLogger(name=__name__)


class PythonGNUPGBackend(GPGBackend):
    @staticmethod
    def _decrypt_file(gpg, file_object, keys):
        for key in keys:
            gpg.import_keys(
                key_data=key['key_data']
            )

        return gpg.decrypt_file(file=file_object)

    @staticmethod
    def _import_and_list_keys(gpg, **kwargs):
        import_results = gpg.import_keys(**kwargs)
        return import_results, gpg.list_keys(
            keys=import_results.fingerprints[0]
        )[0]

    @staticmethod
    def _import_key(gpg, **kwargs):
        return gpg.import_keys(**kwargs)

    @staticmethod
    def _list_keys(gpg, **kwargs):
        return gpg.list_keys(**kwargs)

    @staticmethod
    def _recv_keys(gpg, keyserver, key_id):
        import_results = gpg.recv_keys(keyserver, key_id)
        if import_results.count:
            key_data = gpg.export_keys(
                import_results.fingerprints[0]
            )
        else:
            key_data = None
        return key_data

    @staticmethod
    def _search_keys(gpg, keyserver, query):
        return gpg.search_keys(keyserver=keyserver, query=query)

    @staticmethod
    def _sign_file(
        binary, clearsign, detached, file_object, gpg, key_data, output,
        passphrase
    ):
        import_results = gpg.import_keys(key_data=key_data)

        return gpg.sign_file(
            binary=binary, clearsign=clearsign, detach=detached,
            file=file_object, keyid=import_results.fingerprints[0],
            output=output, passphrase=passphrase
        )

    @staticmethod
    def _verify_file(gpg, file_object, keys, data_filename=None):
        for key in keys:
            gpg.import_keys(
                key_data=key['key_data']
            )

        return gpg.verify_file(
            file=file_object, data_filename=data_filename
        )

    def decrypt_file(self, file_object, keys):
        return self.gpg_command(
            file_object=file_object,
            function=PythonGNUPGBackend._decrypt_file, keys=keys
        )

    def gpg_command(self, function, **kwargs):
        with TemporaryDirectory() as temporary_directory:
            gpg = gnupg.GPG(
                gpgbinary=self.kwargs['gpg_path'],
                gnupghome=temporary_directory
            )
            return function(gpg=gpg, **kwargs)

    def import_and_list_keys(self, key_data):
        return self.gpg_command(
            function=PythonGNUPGBackend._import_and_list_keys,
            key_data=key_data
        )

    def import_key(self, key_data):
        return self.gpg_command(
            function=PythonGNUPGBackend._import_key, key_data=key_data
        )

    def list_keys(self, keys):
        return self.gpg_command(
            function=PythonGNUPGBackend._list_keys, keys=keys
        )

    def recv_keys(self, keyserver, key_id):
        return self.gpg_command(
            function=PythonGNUPGBackend._recv_keys, key_id=key_id,
            keyserver=keyserver
        )

    def search_keys(self, keyserver, query):
        return self.gpg_command(
            function=PythonGNUPGBackend._search_keys, keyserver=keyserver,
            query=query
        )

    def sign_file(
        self, binary, clearsign, detached, file_object, key_data, output,
        passphrase
    ):
        return self.gpg_command(
            binary=binary, clearsign=clearsign, detached=detached,
            file_object=file_object, function=PythonGNUPGBackend._sign_file,
            key_data=key_data, output=output, passphrase=passphrase
        )

    def verify_file(self, file_object, keys, data_filename=None):
        return self.gpg_command(
            data_filename=data_filename, file_object=file_object,
            function=PythonGNUPGBackend._verify_file, keys=keys
        )



import json
import unittest

from django.db import connection, models

from rest_framework import status
from rest_framework.reverse import reverse

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin
from mayan.apps.testing.tests.base import GenericViewTestCase

from .. import generics, serializers

from .base import BaseAPITestCase
from .literals import TEST_OBJECT_LABEL, TEST_OBJECT_LABEL_EDITED
from .mixins import DynamicFieldSerializerAPIViewTestCaseMixin, RESTAPIViewTestMixin


class RESTAPIViewTestCase(RESTAPIViewTestMixin, GenericViewTestCase):
    def test_browser_api_view(self):
        response = self._request_test_browser_api_view()
        self.assertEqual(response.status_code, 200)

    @unittest.skipIf(
        condition=connection.vendor != 'sqlite',
        reason='Skip for known Django issues #15802 and #27074'
    )
    def test_redoc_ui_view(self):
        response = self._request_test_redoc_ui_view()
        self.assertEqual(response.status_code, 200)

    @unittest.skipIf(
        condition=connection.vendor != 'sqlite',
        reason='Skip for known Django issues #15802 and #27074'
    )
    def test_swagger_ui_view(self):
        response = self._request_test_swagger_ui_view()
        self.assertEqual(response.status_code, 200)

    def test_swagger_no_ui_json_view(self):
        self.expected_content_types = ('application/json; charset=utf-8',)

        response = self._request_test_swagger_no_ui_json_view()
        self.assertEqual(response.status_code, 200)

    def test_swagger_no_ui_yaml_view(self):
        self.expected_content_types = ('application/yaml; charset=utf-8',)

        response = self._request_test_swagger_no_ui_yaml_view()
        self.assertEqual(response.status_code, 200)


class BatchAPIRequestViewTestCase(BaseAPITestCase):
    def setUp(self):
        super().setUp()

        self._create_test_permission()

        self.TestModel = self._create_test_model(
            fields={
                'label': models.CharField(max_length=32, unique=True)
            }
        )

        ModelPermission.register(
            model=self.TestModel, permissions=(
                self._test_permission,
            )
        )

        self._create_test_object(
            instance_kwargs={
                'label': TEST_OBJECT_LABEL
            }
        )

        class TestModelSerializer(serializers.ModelSerializer):
            class Meta:
                fields = ('id', 'label')
                model = self.TestModel

        def _test_view_factory():
            class TestView(generics.ListCreateAPIView):
                mayan_object_permissions = {
                    'GET': (self._test_permission,)
                }
                mayan_view_permissions = {
                    'POST': (self._test_permission,)
                }
                serializer_class = TestModelSerializer
                source_queryset = self.TestModel.objects.all()

            return TestView.as_view()

        self.add_test_view(
            test_object=self._test_object,
            test_view_factory=_test_view_factory,
            test_view_url=r'^test-view-url/$'
        )
        self._test_model_list_api_view_name = self._test_view_name

        def _test_view_factory():
            TestModel = self.TestModel

            class TestView(generics.RetrieveUpdateDestroyAPIView):
                lookup_url_kwarg = 'test_object_id'
                mayan_object_permissions = {
                    'DELETE': (self._test_permission,),
                    'GET': (self._test_permission,),
                    'PATCH': (self._test_permission,),
                    'PUT': (self._test_permission,)
                }
                serializer_class = TestModelSerializer
                source_queryset = TestModel.objects.all()

            return TestView.as_view()

        self.add_test_view(
            test_object=self._test_object,
            test_view_factory=_test_view_factory,
            test_view_url=r'^test-view-url/(?P<test_object_id>\d+)/$'
        )
        self._test_model_detail_api_view_name = self._test_view_name

    def _request_batch_api_request_api_view(self, requests):
        return self.post(
            viewname='rest_api:batchrequest-create', data={
                'requests': requests
            }
        )

    def test_create_batch_api_request(self):
        self.grant_permission(permission=self._test_permission)

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL},
                'method': 'POST',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            }
        ]

        self._test_object.delete()

        test_model_count = self.TestModel.objects.count()

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'],
            status.HTTP_201_CREATED
        )

        self.assertEqual(
            self.TestModel.objects.count(), test_model_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_delete_batch_api_request(self):
        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        requests = [
            {
                'method': 'DELETE',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self._test_object.pk}
                )
            }
        ]

        test_model_count = self.TestModel.objects.count()

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'],
            status.HTTP_204_NO_CONTENT
        )

        self.assertEqual(
            self.TestModel.objects.count(), test_model_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_edit_via_patch_batch_api_request(self):
        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'method': 'PATCH',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self._test_object.pk}
                )
            }
        ]

        test_model_label = self._test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )

        self._test_object.refresh_from_db()
        self.assertNotEqual(self._test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_edit_via_put_batch_api_request(self):
        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'method': 'PUT',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self._test_object.pk}
                )
            }
        ]

        test_model_label = self._test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )

        self._test_object.refresh_from_db()
        self.assertNotEqual(self._test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_list_get_batch_api_request(self):
        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        requests = [
            {
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            }
        ]

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )
        self.assertEqual(response.data['results'][0]['data']['count'], 1)
        self.assertEqual(
            response.data['results'][0]['data']['results'][0]['id'],
            self._test_object.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mass_edit_get_batch_api_request(self):
        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        requests = [
            {
                'name': 'test_object_list',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            },
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'iterables': ['test_object_list.data.results'],
                'method': 'PATCH',
                'name': 'test_object_edit',
                'url': '{}{{{{ iterables.0.id }}}}/'.format(
                    reverse(
                        viewname='rest_api:{}'.format(
                            self._test_model_list_api_view_name
                        )
                    )
                )
            }
        ]

        test_model_label = self._test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 2)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )
        self.assertEqual(response.data['results'][0]['data']['count'], 1)
        self.assertEqual(
            response.data['results'][0]['data']['results'][0]['id'],
            self._test_object.pk
        )

        self.assertEqual(
            response.data['results'][1]['status_code'], status.HTTP_200_OK
        )

        self._test_object.refresh_from_db()
        self.assertNotEqual(self._test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DynamicFieldSerializerAPIViewTestCase(
    DynamicFieldSerializerAPIViewTestCaseMixin, BaseAPITestCase
):
    def _get_test_view_class(self, serializer_class):
        local_serializer_class = serializer_class

        class TestView(generics.RetrieveAPIView):
            lookup_url_kwarg = 'test_object_id'
            serializer_class = local_serializer_class
            source_queryset = self.TestModelChild.objects.all()

        return TestView

    def test_current_model_only_field_single(self):
        response = self._request_test_api_view(
            query={'_fields_only': 'test_field_3'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' not in data)
        self.assertTrue('test_field_3' in data)
        self.assertTrue('test_field_4' not in data)

    def test_current_model_only_field_multiple(self):
        response = self._request_test_api_view(
            query={'_fields_only': 'test_field_3,test_field_4'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' not in data)
        self.assertTrue('test_field_3' in data)
        self.assertTrue('test_field_4' in data)

    def test_current_model_only_rleated_field(self):
        response = self._request_test_api_view(
            query={'_fields_only': 'parent'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' not in data)

    def test_related_model_only_field_single(self):
        response = self._request_test_api_view(
            query={'_fields_only': 'parent__test_field_1'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' not in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' not in data)

    def test_related_model_only_field_multiple(self):
        response = self._request_test_api_view(
            query={
                '_fields_only': 'parent__test_field_1,parent__test_field_2'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' not in data)

    def test_current_model_exclude_field_single(self):
        response = self._request_test_api_view(
            query={'_fields_exclude': 'test_field_3'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' in data)

    def test_current_model_exclude_field_multiple(self):
        response = self._request_test_api_view(
            query={'_fields_exclude': 'test_field_3,test_field_4'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' not in data)

    def test_current_model_exclude_related_field(self):
        response = self._request_test_api_view(
            query={'_fields_exclude': 'parent'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' not in data)
        self.assertTrue('test_field_3' in data)
        self.assertTrue('test_field_4' in data)

    def test_related_model_exclude_field_single(self):
        response = self._request_test_api_view(
            query={'_fields_exclude': 'parent__test_field_1'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' not in data['parent'])
        self.assertTrue('test_field_2' in data['parent'])
        self.assertTrue('test_field_3' in data)
        self.assertTrue('test_field_4' in data)

    def test_related_model_exclude_field_multiple(self):
        response = self._request_test_api_view(
            query={
                '_fields_exclude': 'parent__test_field_1,parent__test_field_2'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' not in data['parent'])
        self.assertTrue('test_field_2' not in data['parent'])
        self.assertTrue('test_field_3' in data)
        self.assertTrue('test_field_4' in data)


class DynamicFieldSerializerWithMixinAPIViewTestCase(
    DynamicFieldSerializerAPIViewTestCaseMixin, BaseAPITestCase
):
    auto_add_test_view = True
    auto_create_test_object = False
    test_view_url = r'^test-view-url/(?P<test_object_id>\d+)/$'

    def _get_test_view_class(self, serializer_class):
        local_serializer_class = serializer_class

        class TestView(ExternalObjectAPIViewMixin, generics.RetrieveAPIView):
            external_object_queryset = self.TestModelChild.objects.all()
            external_object_pk_url_kwarg = 'test_object_id'
            lookup_url_kwarg = 'test_object_id'
            serializer_class = local_serializer_class
            source_queryset = self.TestModelChild.objects.all()

        return TestView

    def test_related_model_only_field_single_with_api_view_mixins(self):
        response = self._request_test_api_view(
            query={'_fields_only': 'parent__test_field_1'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertTrue('parent' in data)
        self.assertTrue('test_field_1' in data['parent'])
        self.assertTrue('test_field_2' not in data['parent'])
        self.assertTrue('test_field_3' not in data)
        self.assertTrue('test_field_4' not in data)



from django.utils.translation import ugettext_lazy as _


class CheckedOutDocumentBusinessLogicMixin:
    def get_user_display(self):
        check_out_info = self.get_check_out_info()
        return check_out_info.user.get_full_name() or check_out_info.user

    get_user_display.short_description = _('User')

    def get_checkout_datetime(self):
        return self.get_check_out_info().checkout_datetime

    get_checkout_datetime.short_description = _('Checkout time and date')

    def get_checkout_expiration(self):
        return self.get_check_out_info().expiration_datetime

    get_checkout_expiration.short_description = _('Checkout expiration')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Document downloads'), name='document_downloads'
)

permission_document_file_download = namespace.add_permission(
    label=_('Download document files'),
    name='document_file_download'
)



from mayan.apps.documents.search import search_model_document

from ...literals import QUERY_PARAMETER_ANY_FIELD, SEARCH_MODEL_NAME_KWARG


class SearchAPIViewTestMixin:
    def _request_search_advanced_view(
        self, search_model_name=None, search_term=None, query=None
    ):
        if search_term is None:
            search_term = self._test_document.document_type.label

        view_query = {
            'document_type__label': search_term
        }
        if query:
            view_query.update(query)

        search_model_name = search_model_name or search_model_document.full_name

        return self.get(
            viewname='rest_api:advanced-search-view', kwargs={
                SEARCH_MODEL_NAME_KWARG: search_model_name
            }, query=view_query
        )

    def _request_search_simple_view(
        self, search_model_name=None, search_term=None, query=None
    ):
        if search_term is None:
            search_term = self._test_document.label

        view_query = {
            QUERY_PARAMETER_ANY_FIELD: search_term
        }
        search_model_name = search_model_name or search_model_document.full_name

        if query:
            view_query.update(query)

        return self.get(
            viewname='rest_api:search-view', kwargs={
                SEARCH_MODEL_NAME_KWARG: search_model_name
            }, query=view_query
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_signature_capture_create, icon_signature_capture_single_delete,
    icon_signature_capture_edit, icon_signature_capture_list
)
from .permissions import (
    permission_signature_capture_create, permission_signature_capture_delete,
    permission_signature_capture_edit, permission_signature_capture_view
)

link_signature_capture_create = Link(
    args='object.pk', icon=icon_signature_capture_create, permissions=(
        permission_signature_capture_create,
    ), text=_('Create new signature capture'),
    view='signature_captures:signature_capture_create'
)
link_signature_capture_delete = Link(
    args='object.pk', icon=icon_signature_capture_single_delete,
    permissions=(permission_signature_capture_delete,), tags='dangerous',
    text=_('Delete'),
    view='signature_captures:signature_capture_delete'
)
link_signature_capture_edit = Link(
    args='object.id', icon=icon_signature_capture_edit,
    permissions=(permission_signature_capture_edit,), text=_('Edit'),
    view='signature_captures:signature_capture_edit'
)
link_signature_capture_list = Link(
    args='resolved_object.pk', icon=icon_signature_capture_list,
    permissions=(permission_signature_capture_view,), text=_(
        'Signature captures'
    ), view='signature_captures:signature_capture_list'
)
